@@ -2,6 +2,7 @@ import datetime
 import io
 import itertools
 import logging
+import ssl
 import threading
 import time
 from collections import defaultdict
@@ -18,17 +19,40 @@ logger = logging.getLogger('echolog')
 DEFAULT_AE_TITLE = "Us2.ai"
 
 
+def server_context(ca, cert, key):
+    context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
+    context.verify_mode = ssl.CERT_REQUIRED
+    context.load_cert_chain(certfile=cert, keyfile=key)
+    context.load_verify_locations(cafile=ca)
+    # Only TLS <= 1.2 is supported, make sure we always use this
+    context.minimum_version = context.maximum_version = ssl.TLSVersion.TLSv1_2
+
+    return context
+
+
+def client_context(cert, key):
+    context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH, cafile=cert)
+    context.verify_mode = ssl.CERT_OPTIONAL
+    context.load_cert_chain(certfile=cert, keyfile=key)
+    context.check_hostname = False
+
+    return context
+
+
 class PacsConnection:
     def __init__(self, details):
         parts = details.split(':')
         self.host, port, self.remote_ae_title = parts[:3]
         self.local_ae_title = parts[3] if len(parts) > 3 else DEFAULT_AE_TITLE
+        self.cert, self.key = parts[4:6] if len(parts) > 5 else (None, None)
         self.port = int(port)
 
     def store(self, ds):
         ae = AE(ae_title=self.local_ae_title)
         ae.add_requested_context(ds.SOPClassUID, ds.file_meta.TransferSyntaxUID)
-        assoc = ae.associate(self.host, self.port, ae_title=self.remote_ae_title)
+        assoc = ae.associate(self.host, self.port, ae_title=self.remote_ae_title,
+                             tls_args=(client_context(self.cert, self.key), None) if self.cert else None)
+
         if not assoc.is_established:
             raise ConnectionError('Association rejected, aborted or never connected')
         # Use the C-STORE service to send the dataset
@@ -65,10 +89,11 @@ class Sync(threading.Thread):
         self.uploader = self.auth.username
         self.killed = False
         self.params = {'v': '1.4.3'}
-        self.last_sync = datetime.datetime.min.replace(tzinfo=datetime.timezone.utc)
+        self.last_sync = datetime.datetime.now().replace(tzinfo=datetime.timezone.utc)
         self.modalities = cmd.sync_modalities
         self.poll = cmd.sync_poll
         self.sr_params = {}
+        self.doc_params = {}
         self.pool = pool
         self.search_params = {k: v for e in cmd.sync_search for k, v in [e.split('=', 1)]}
         self.protocol = unpack(requests.get(
@@ -77,8 +102,13 @@ class Sync(threading.Thread):
             self.sr_params['url'] = True
         if cmd.sync_main_findings:
             self.sr_params['main_findings'] = True
+            self.doc_params['main_findings'] = True
+        if cmd.sync_pdf_images:
+            self.doc_params['image'] = True
         if cmd.sync_designators:
             self.sr_params['designators'] = cmd.sync_designators
+
+        self.doc_params['dicom_encapsulated_pdf'] = True
         self.ps_params = {}
         self.sc_params = {}
 
@@ -127,6 +157,7 @@ class SyncWorker:
         self.auth = worker.auth
         self.params = worker.params
         self.sr_params = worker.sr_params
+        self.doc_params = worker.doc_params
         self.ps_params = worker.ps_params
         self.sc_params = worker.sc_params
         self.protocol = worker.protocol
@@ -148,6 +179,10 @@ class SyncWorker:
         return requests.get(f"{self.api_url}/dicom/sc", headers=self.auth.get_headers(),
                             params={**self.params, **self.sc_params, 'measurements': ms})
 
+    def doc(self):
+        return requests.get(f"{self.api_url}/study/pdf/{self.sid}", headers=self.auth.get_headers(),
+                            params={**self.params, **self.doc_params})
+
     @cached_property
     def mods(self):
         params = {**self.params, 'page': 1, 'page_size': 10_000}
@@ -190,6 +225,7 @@ class SyncWorker:
         for req in itertools.chain(
                 self.media() if [m in self.modalities for m in ['PS', 'SC']] else [],
                 [self.sr()] if 'SR' in self.modalities else [],
+                [self.doc()] if 'DOC' in self.modalities else [],
         ):
             url = req.url
             try:
