@@ -1,11 +1,15 @@
 import argparse
 import base64
 import csv
+import datetime
+import hashlib
 import io
+import itertools
 import json
 import logging
 import math
 import os
+import re
 import socket
 import sys
 import threading
@@ -24,22 +28,25 @@ import pydicom
 import requests
 from pathy import Pathy
 from pydicom import uid
+from pydicom._uid_dict import UID_dictionary
+from pydicom.encaps import decode_data_sequence
 from pydicom.encaps import encapsulate, generate_pixel_data_frame
 from pydicom.errors import InvalidDicomError
 from pydicom.filewriter import write_file_meta_info
 from pydicom.pixel_data_handlers import apply_color_lut
-from pynetdicom import AE, evt, ALL_TRANSFER_SYNTAXES
-from pynetdicom import sop_class
+from pynetdicom import AE, evt, ALL_TRANSFER_SYNTAXES, sop_class
 from tqdm import tqdm
 from watchdog.events import FileCreatedEvent, FileMovedEvent, FileModifiedEvent
 from watchdog.events import FileSystemEventHandler
 from watchdog.observers.polling import PollingObserver
 
 from echoloader.lib import unpack
-from echoloader.sync import PacsConnection, Sync, DEFAULT_AE_TITLE
+from echoloader.sync import PacsConnection, Sync, DEFAULT_AE_TITLE, server_context
 
 logger = logging.getLogger('echolog')
 OS_VAR_PREFIX = 'US2_'
+for k in ['1.2.840.10008.5.1.4.1.1.3', '1.2.840.10008.5.1.4.1.1.6']:
+    sop_class._STORAGE_CLASSES[UID_dictionary[k][-1]] = k
 
 
 def setup_logging(filename=None, verbose=None):
@@ -164,6 +171,7 @@ def blank_top_bar(media, regions):
         yo, yn = region.RegionLocationMinY0, region.RegionLocationMaxY1
         mask[yo:yn, xo:xn] = 1
     media *= mask
+    return media
 
 
 def mpeg4hp41(ds):
@@ -171,33 +179,45 @@ def mpeg4hp41(ds):
     return np.asarray(frames)
 
 
-def unusual_frame_mean(px: np.ndarray, threshold=58):
+def unusual_frame_mean(px: np.ndarray, threshold=100):
     """
     If mean pixel value of frame is larger than threshold, background is likely non-black
     (usually happens when dicom tag is RGB but the frames are actually in YBR).
     """
     frame = px[px.shape[0] // 2] if is_video(px) else px  # take middle frame for video
-    return frame.mean() > threshold
+    return cv2.cvtColor(frame, cv2.COLOR_BGR2HSV)[..., -1].mean() > threshold
 
 
-def parse_dicom_pixel(dicom):
-    """Parse color space and coerce to RGB, and anonymize by blanking out top bar."""
+def media_iter(dicom):
     try:
-        px = dicom.pixel_array
-    except NotImplementedError as exc:
+        seq = decode_data_sequence(dicom.PixelData)
+        if not seq:
+            raise ValueError('No data sequence found')
+        for b in seq:
+            yield imageio.imread(io.BytesIO(b))
+    except Exception as exc:
+        logger.info(f'Failed to decode JPEG compressed frame bytes due to {exc}, falling back to pixel_array')
         handler = parse_dicom_pixel.custom_handlers.get(dicom.file_meta.TransferSyntaxUID)
-        if not handler:
-            raise exc
-        px = handler(dicom)
+        px = handler(dicom) if handler else dicom.pixel_array
+        if isinstance(px, np.ndarray) and not is_video(px):
+            px = [px]
+        yield from px
+
+
+def parse_dicom_pixel(dicom, anonymize=True):
+    """Parse color space and coerce to RGB, and anonymize by blanking out top bar."""
+    px = media_iter(dicom)
     pi = dicom.PhotometricInterpretation
-    dicom.PhotometricInterpretation = 'RGB'
-    if pi in ['YBR_FULL', 'YBR_FULL_422', 'RGB'] and unusual_frame_mean(px):
-        px = np.asarray([ybr_to_rgb(img) for img in px]) if is_video(px) else ybr_to_rgb(px)
-    elif pi in ['PALETTE COLOR']:
-        px = (apply_color_lut(px, dicom) // 255).astype('uint8')
-    else:
-        dicom.PhotometricInterpretation = pi
-    return px
+    px = (ybr_to_rgb(img) if pi in ['YBR_FULL', 'YBR_FULL_422', 'RGB'] and unusual_frame_mean(img) else img
+          for img in px)
+    px = ((apply_color_lut(img, dicom) // 255).astype('uint8') if pi in ['PALETTE COLOR'] else img for img in px)
+    px = (np.repeat(np.expand_dims(img, -1), 3, -1) if len(img.shape) < 3 else img for img in px)
+    if anonymize:
+        px = (blank_top_bar(img, getattr(dicom, "SequenceOfUltrasoundRegions", [])) for img in px)
+    sample = next(px)
+    px = itertools.chain([sample], px)
+    shape = (getattr(dicom, 'NumberOfFrames', 1), *sample.shape)
+    return px, shape
 
 
 parse_dicom_pixel.custom_handlers = {
@@ -227,25 +247,20 @@ def package_dicom(ds, anonymize, compress):
     ds.remove_private_tags()
     if not anonymize and not compress:
         return
-    media = parse_dicom_pixel(ds)
+    media, shape = parse_dicom_pixel(ds, anonymize)
     if anonymize:
         ds.walk(person_data_callback)
-        blank_top_bar(media, getattr(ds, "SequenceOfUltrasoundRegions", []))
-    video = is_video(media)
-
     ds.is_little_endian = True
     ds.is_implicit_VR = False
 
     ds.BitsStored = 8
     ds.BitsAllocated = 8
     ds.HighBit = 7
-    if len(media.shape) < 3 + video:
-        media = np.repeat(np.expand_dims(media, -1), 3, -1)
-    ds.Rows, ds.Columns, ds.SamplesPerPixel = media.shape[video:]
+    ds.Rows, ds.Columns, ds.SamplesPerPixel = shape[1:]
     ds.PhotometricInterpretation = "YBR_FULL_422"
-    if video:
+    if shape[0] > 1:
         ds.StartTrim = 1
-        ds.StopTrim = ds.NumberOfFrames = media.shape[0] if video else 1
+        ds.StopTrim = ds.NumberOfFrames = shape[0]
         fps = getattr(ds, 'CineRate', 1000 / getattr(ds, 'FrameTime', 40))
         ds.CineRate = ds.RecommendedDisplayFrameRate = round(fps, 2)
         ds.FrameTime = 1000 / ds.CineRate
@@ -255,13 +270,14 @@ def package_dicom(ds, anonymize, compress):
         if compress:
             ds.file_meta.TransferSyntaxUID = uid.MPEG4HP41
             ds.PixelData = encapsulate([
-                imageio.mimwrite(imageio.RETURN_BYTES, pad_to_multiple(media, 16), fps=fps, format='mp4')])
+                imageio.mimwrite(imageio.RETURN_BYTES, (pad_to_multiple(img, 16, (0, 1)) for img in media), fps=fps,
+                                 format='mp4')])
         else:
             ds.file_meta.TransferSyntaxUID = uid.JPEGBaseline8Bit
             ds.PixelData = encapsulate([imageio.imwrite(imageio.RETURN_BYTES, img, format='jpg') for img in media])
     else:
         ds.file_meta.TransferSyntaxUID = uid.JPEGBaseline8Bit
-        ds.PixelData = encapsulate([imageio.imwrite(imageio.RETURN_BYTES, media, format='jpg')])
+        ds.PixelData = encapsulate([imageio.imwrite(imageio.RETURN_BYTES, next(media), format='jpg')])
     ds['PixelData'].is_undefined_length = True
 
 
@@ -371,8 +387,9 @@ class Handler(FileSystemEventHandler):
 
 
 class HandlerWorker:
-    def __init__(self, handler: Handler, path=None, ds=None, ae_title=None, f=None):
+    def __init__(self, handler: Handler, path=None, ds=None, ae_title=None, f=None, **kwargs):
         self.path = path
+        self.kwargs = kwargs
         self.ds = ds
         self.ae_title = ae_title
         self.f = f
@@ -381,12 +398,13 @@ class HandlerWorker:
     def file_params(self, ds, ae_title, filename=None):
         customer = getattr(self.args, 'customer', '') or ae_title
         stem = '_'.join(map(str, filter(bool, [ds.SOPInstanceUID, filename])))
+        kwargs = self.kwargs
         return {
             'customer': customer or '',
-            'trial': getattr(self.args, 'trial', None) or customer or '',
-            'patient_id': ds.PatientID,
-            'visit_id': ds.StudyID or ds.StudyInstanceUID or ds.StudyDate or 'No Study ID',
-            'filename': f"{stem}.dcm",
+            'trial': kwargs.get('trial') or getattr(self.args, 'trial', None) or customer or '',
+            'patient_id': kwargs.get('patient_id') or ds.PatientID,
+            'visit_id': kwargs.get('visit_id') or ds.StudyID or ds.StudyInstanceUID or ds.StudyDate or 'No Study ID',
+            'filename': kwargs.get('filename') or f"{stem}.dcm",
         }
 
     def upload(self, ds, param):
@@ -410,17 +428,31 @@ class HandlerWorker:
         buf.seek(0)
         return unpack(requests.put(url, data=buf.read(), headers=headers, params=upload_param))
 
+    def pseudonymize(self, ds):
+        for tag in self.args.pseudonymize:
+            old = getattr(ds, tag)
+            new = hashlib.sha256(old.encode()).hexdigest()[:8]
+            logger.info(f'Pseudonymizing {tag} from {getattr(ds, tag)} to {new}')
+            if self.args.pseudonymize_file:
+                f = Pathy.fluid(self.args.pseudonymize_file).open('a+')
+                f.seek(0)
+                entry = f"{tag},{old},{new}"
+                if entry not in set(f):
+                    f.write(f"{entry}\n")
+            setattr(ds, tag, new)
+
     def process(self):
         try:
             path = self.path and Pathy.fluid(self.path)
             ds = self.ds or pydicom.dcmread(self.f or path.open('rb'), force=True)
+            if self.args.pseudonymize:
+                self.pseudonymize(ds)
             params = self.file_params(ds, ae_title=self.ae_title, filename=getattr(path, 'stem', None))
             logger.info(f'processing {params}')
             extracted = self.args.extracted
             k = tuple(params.values())
             if k not in extracted:
-                if self.args.filter and not any(
-                        getattr(ds, k, None) == v for f in self.args.filter for k, v in [f.split('=', 1)]):
+                if self.args.filter and not all(eval(f, {'ds': ds}) for f in self.args.filter):
                     logger.info(f'Skipping {ds.SOPInstanceUID} due to no matching filter')
                     return
                 package_dicom(ds, anonymize=self.args.anonymize, compress=self.args.compress)
@@ -437,6 +469,17 @@ class HandlerWorker:
                 extracted.add(k)
                 if self.args.csv_out:
                     csv.writer(open(self.args.csv_out, 'a', newline='')).writerow(k)
+        except Exception as exc:
+            if self.args.save_failed and self.f:
+                try:
+                    self.f.seek(0)
+                    self.args.save_failed.mkdir(parents=True, exist_ok=True)
+                    dst = Pathy.fluid(self.args.save_failed) / f'{datetime.datetime.now().strftime("%m%d%Y%H%M%S")}.dcm'
+                    dst.open('wb').write(self.f.read())
+                except Exception as e:
+                    logger.error(f'Error saving failed: {e}')
+                    logger.debug(traceback.format_tb(e.__traceback__))
+            raise exc
         finally:
             if self.f:
                 self.f.close()
@@ -480,6 +523,11 @@ def read_args():
     parser.add_argument(
         "--src",
         help="The folder to anonymize")
+    parser.add_argument(
+        "--src-details",
+        help="Regular expression to apply on path to extract patient details used in the upload request. "
+             "Can use (?P<key>re) to capture regular expression re to key. "
+             "Key can be trial, patient_id, visit_id or filename")
     parser.add_argument(
         "--dst",
         help="The output folder for the anonymized DICOM, defaults to src folder suffixed with '_anonymized'")
@@ -489,17 +537,16 @@ def read_args():
     parser.add_argument(
         "--pacs", action="store_true",
         help="Starts PACS server")
+    parser.add_argument(
+        "--secure", nargs=3,
+        help="Enable Secure PACS, must include 3 arguments which are paths to: "
+             "CA certificate, server certificate and server key")
     parser.add_argument(
         "--pacs-ae-title", default=DEFAULT_AE_TITLE,
         help="PACS AE Title, defaults to Us2.ai")
     parser.add_argument(
         "--pacs-port", default=11112, type=int,
         help="PACS port, defaults to 11112")
-    parser.add_argument(
-        "--pacs-sops", type=str, nargs='+', default={
-            sop for sop in vars(sop_class).values()
-            if isinstance(sop, str) and sop.count('.') >= 5 and all(s.isdigit() for s in sop.split('.'))},
-        help="PACS accepted SOPs")
     parser.add_argument(
         "--overwrite", action="store_true",
         help="Overwrite files in the output folder")
@@ -521,10 +568,21 @@ def read_args():
     parser.add_argument(
         "--verbose", action='store_true',
         help="Enable all application debug/info logs")
+    parser.add_argument(
+        "--save-failed", type=Pathy.fluid,
+        help="Path to store failed cases locally")
     parser.add_argument(
         '--no-anonymization', action='store_false', dest='anonymize',
         help="No anonymization of data prior to upload",
     )
+    parser.add_argument(
+        '--pseudonymize', nargs='+',
+        help="DCM tags to pseudonymize",
+    )
+    parser.add_argument(
+        '--pseudonymize-file',
+        help="Path to file that will be output with mapping of original to pseudonymized values",
+    )
     parser.add_argument(
         '--no-compression', action='store_false', dest='compress',
         help="Don't compress videos into mp4 format",
@@ -537,7 +595,8 @@ def read_args():
         '--sync',
         nargs='+',
         default=[],
-        help="Will try to sync structured report back to specified location. Format: IP:PORT:AE_TITLE",
+        help="Will try to sync structured report back to specified location."
+             "Format: IP:PORT:REMOTE_AE_TITLE[:LOCAL_AE_TITLE:CERT:KEY]",
     )
     parser.add_argument(
         '--sync-url',
@@ -549,7 +608,13 @@ def read_args():
         '--sync-main-findings',
         default=False,
         action='store_true',
-        help='Include main findings in SR export',
+        help='Include main findings in SR export/DICOM encapsulated PDF',
+    )
+    parser.add_argument(
+        '--sync-pdf-images',
+        default=False,
+        action='store_true',
+        help='Include images in DICOM encapsulated PDF',
     )
     parser.add_argument(
         '--sync-designators',
@@ -647,8 +712,21 @@ def main():
             paths = [src] if src.is_file() else [p for p in src.rglob(args.src_glob) if p.is_file()]
             args.n = len(paths)
             args.i = 0
-            for path in paths:
-                handler.handle(path)
+            details = args.src_details and re.compile(args.src_details)
+            for i, path in enumerate(paths):
+                kwargs = {}
+                if details:
+                    match = details.search(path.as_posix())
+                    if match:
+                        kwargs = match.groupdict()
+                        s = f"Matched path details: {kwargs} on {path.as_posix()}"
+                        logger.info(s)
+                    else:
+                        s = f"No match for path details regex on {path.as_posix()}"
+                        logger.warning(s)
+                    if i == 0 and input(f'{s}, Continue (y/n): ') != 'y':
+                        return
+                handler.handle(path, **kwargs)
             if args.watch:
                 logger.warning(f"watching folder {os.path.abspath(src)}")
                 src.mkdir(exist_ok=True, parents=True)
@@ -663,10 +741,14 @@ def main():
                 f"Starting pacs server on {get_ip()}:{args.pacs_port} with AE title {args.pacs_ae_title}")
             handlers = [(evt.EVT_C_STORE, handle_store, [handler])]
             ae = AE(args.pacs_ae_title)
-            for sop in args.pacs_sops:
+            for sop in UID_dictionary.keys():
                 ae.add_supported_context(sop, ALL_TRANSFER_SYNTAXES)
             ae.maximum_pdu_size = 0
-            pacs = ae.start_server(('0.0.0.0', args.pacs_port), block=False, evt_handlers=handlers)
+            pacs = ae.start_server(
+                ('0.0.0.0', args.pacs_port),
+                block=False,
+                ssl_context=server_context(*args.secure) if args.secure else None,
+                evt_handlers=handlers)
         if args.proxy:
             httpd = ThreadingHTTPServer(server_address=('0.0.0.0', args.proxy), RequestHandlerClass=create_proxy(args))
             thread = threading.Thread(target=httpd.serve_forever)
